// solarRegulator sketch
// June 2025  modified version for LGTF328P SSOP20 board
//
// for Mike Payne's solar powered plane project
// by ceptimus. New Year's Day 2018.
// throttle channel input from receiver on 2
// throttle output to ESC on pin 12

// solar cell input voltage into A5 via potential divider
// V.solar ----/\/\/\/\----+----/\/\/\/\---- GND
//                R1       |       R2
//                        A5
// Recommended values:
//                     R1 = 68K
//                     R2 = 10K
// Gives theoretical input range of 2.048 * (68 + 10) / 10 = 15.97V  good for up to 26 solar cells.  Adjust R1 and/or R2 if more cells used
// 15.97V range also allows 2s or 3s LiPo battery to be used for tests (fully charged 3-cell is 12.6V)

#define R1 68000.0
#define R2 10000.0

#define SMOOTHED_ADC_COUNT_TO_VOLTS (2.048 * (R1 + R2) / R2 / 131040.0)

// ramp up rate for increasing throttle output to ESC.  Lower numbers give a faster rate.  1 is the lowest valid value.  2 is twice as slow, 3 three times as slow and so on.  A value of 0 gives an effective value of 256.  2 is a good starting point
#define RAMP_UP_RATE 2

// ramp down rate for decreasing throttle output to ESC when available voltage is too low.  Higher numbers give a faster rate.  100.0 is a good starting point
#define RAMP_DOWN_RATE 100.0

// 1.8V specified minimum for chip + 1.2V dropout for the AMS1117 LDO regulator
#define OPERATING_VOLTAGE 3.0

const float mptRatio = 0.85;  // maximum power transfer ratio.  This is multiplied by the highest (minimum load) voltage seen to give throttleBackVoltage.  85% is said to be optimum.

uint32_t filteredA5; // remove noise from analog inputs by simple 1st order low pass filters ("leaky integrators")

volatile uint16_t rxThrottlePulseWidth = 0; // this will be updated by the interrupt routine when a signal is received from the receiver. 0 is a flag for 'not received yet'
uint16_t minRxThrottlePulseWidth = 2200;
uint16_t maxRxThrottlePulseWidth = 800; // these will keep track of the actual (within acceptable range) minimum and maximum values seen arriving from the receiver
uint16_t maxEscPulseWidth = 0; // this value is updated by the program so as to always keep the solarCellVoltage at or above the selected minimum.  0 is a flag for 'not calculated yet'
float throttleBackVoltage = OPERATING_VOLTAGE; // minimum voltage (to allow Arduino and Rx to always be powered)  Adapts to fraction of highest (minimum load) voltage seen via A3.  85% is reckoned to give max power transfer
float maximumVoltageSeen = OPERATING_VOLTAGE / mptRatio;

void setup() {
  Serial.begin(115200);
  analogReference(INTERNAL2V048);
  analogReadResolution(12);
  Serial.begin(115200);
  pinMode(2, INPUT); // throttle input from receiver
  attachInterrupt(digitalPinToInterrupt(2), pinchange, CHANGE);
  digitalWrite(12, LOW);
  pinMode(12, OUTPUT); // throttle output to ESC
  TCCR1A = 0x00; // intialize timer counter 1 ready to run in normal mode where it counts up and can generate an interrupt when it wraps from 0xFFFF to 0x0000 counts
  TCCR1C = 0x00;

  // enable pinChange interrupt on pin 2
  PCMSK2 = 0x04; // mask bits of PORTD (pin 0 - pin 7) so that only A1 generate pinchange interrupts
  // PCICR |= 0x04; // enable pinchange interrupts for PORTD

  filteredA5 = 0;  // start A5 filter on minimum so that any start-up noise doesn't cause maximumVoltageSeen to be set artificially high
}

void loop() {
  static uint8_t rampUpRate = RAMP_UP_RATE;

  // update analog input filtered value - this is bit-shifted 4 bits so have range 0 - 4095 * 2^5 = 0 - 131040
  filteredA5 -= filteredA5 >> 5;
  filteredA5 += analogRead(A5);

  float solarCellVoltage = filteredA5 * SMOOTHED_ADC_COUNT_TO_VOLTS; // measured solar cell voltage
  if ((millis() < 10000UL) && (solarCellVoltage > maximumVoltageSeen)) { // ignore surges in voltage, such as can be caused by sudden throttling back, after first 10 seconds of operation
    maximumVoltageSeen = solarCellVoltage;
    throttleBackVoltage = mptRatio * maximumVoltageSeen;
  }

  // when within-acceptable-range pulse widths arrive from the receiver, keep track of the minimum and maximum values seen so far
  uint16_t rpw = rxThrottlePulseWidth;
  if (rpw <= 2200 && rpw > maxRxThrottlePulseWidth) {
    maxRxThrottlePulseWidth = rpw;
  }
  if (rpw >= 800 && rpw < minRxThrottlePulseWidth) {
    minRxThrottlePulseWidth = rpw;
  }

  if (maxRxThrottlePulseWidth >= minRxThrottlePulseWidth) { // don't alter maxEscPulseWidth until we've begun receiving throttle input from the receiver
    uint16_t mpw = maxEscPulseWidth;
    if (solarCellVoltage <= throttleBackVoltage) { // need to reduce maxEscPulseWidth so as to reduce power draw on solar cells
      mpw -= (uint16_t)((throttleBackVoltage - solarCellVoltage) * RAMP_DOWN_RATE); // ramp power down pretty quickly (this is a potential emergency!) but proportional to the severity of the under voltage condition
    } else if (rpw > maxEscPulseWidth) { // extra power is available and user is requesting more power than maxEscPulseWidth currently permits
      if (!--rampUpRate) {
        ++mpw; // ramp up the ESC signal fairly slowly
        rampUpRate = RAMP_UP_RATE;
      }
    } else { // user is requesting less power than maxEscPulseWidth
      mpw = rpw; // set maxEscPulseWidth to current throttle position.  This is necessary to prevent fast ramping up of power next time the user increases the throttle - lighting conditions may have changed and a fast ramp up may trip out the ESC
    }
    if (mpw < minRxThrottlePulseWidth) {
      mpw = minRxThrottlePulseWidth;
    } else if (mpw > maxRxThrottlePulseWidth) {
      mpw = maxRxThrottlePulseWidth;
    }
    if (mpw != maxEscPulseWidth) {
      noInterrupts(); // altering a 16-bit value used by the pinchange/timer interrupts so disable interrupts...
      maxEscPulseWidth = mpw; // ...while updating the value to prevent possible glitches when the value changes from, say, 1791 to 1792 (0x05FF to 0x0600)
      interrupts();
    }
  }

  // probably best to leave these debugging print commands in place, even if you're not using them, as they affect the loop time
  Serial.print(solarCellVoltage);
  Serial.print("V\t");
  Serial.print(maximumVoltageSeen);
  Serial.print("V\t");
  Serial.print(throttleBackVoltage);
  Serial.print("V\t");
  Serial.print(rxThrottlePulseWidth);
  Serial.print("us\t");
  Serial.print(maxEscPulseWidth);
  Serial.print("us\r\n");
}

void pinchange(void) { // pinchange occured on pin 2
  static uint16_t pulseRisingEdgeTime;
  if (PIND & 0x04) { // positive edge of PWM signal from receiver
    PORTB |= 0x10; // echo positive edge out to ESC on pin 12
    if (maxEscPulseWidth) { // if main loop has seen valid receiver signals and calculated a maximum ESC pulse width
      pulseRisingEdgeTime = ~(maxEscPulseWidth << 2); // when running with 32MHz internal oscillator, timer runs at 4 MHz
    } else {
      pulseRisingEdgeTime = 0x0001;
    }

    TCNT1 = pulseRisingEdgeTime;
    TIMSK1 = 0x01; // enable timer interrupt
    TCCR1B = 0x02; // start timer running at 4MHz (32MHz internal oscillator) - it will generate an interrupt after maxEscPulseWidth microsecond
  } else { // negative edge of PWM signal from receiver
    PORTB &= ~0x10; // echo negative edge out to ESC on pin 12 (unless the timer limiting the pulse width to maxEscPulseWidth occurred first)
    uint16_t pulseFallingEdgeTime = TCNT1;
    TCCR1B = 0x00; // stop timer1 now until we need it again
    rxThrottlePulseWidth =  (pulseFallingEdgeTime - pulseRisingEdgeTime) >> 2; // when running with a 32MHz internal oscillatorl, divide counter by 4 to get microseconds
  }
}

ISR(TIMER1_OVF_vect) { // timer1 has triggered after maxEscPulseWidth microseconds
  TIMSK1 = 0; // prevent further interrupts from timer/counter 1 till reenabled
  PORTB &= ~0x10; // send negative edge out to ESC on pin 12 (unless the falling edge from the receiver signal occurred first)
}
